/***
CHIPSEC: Platform Security Assessment Framework
Copyright (c) 2023, Intel Corporation


This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; Version 2.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

Contact information:
chipsec@intel.com
*/


#ifndef _FILTDEBUG__H
#define _FILTDEBUG__H

#ifdef __cplusplus
extern "C"
{
#endif

#define DEBUGPRINT 0    // Set to 1 to do DebugPrints in free version

#if DBG || DEBUGPRINT
#define DODEBUGPRINT 1
#else
#define DODEBUGPRINT 0
#endif

#if DODEBUGPRINT
    void DebugPrintInit(char*);
    void DebugPrintClose();
    void DebugPrintMsg(char*);
#else

#define DebugPrintInit(x)
#define DebugPrintClose()
#define DebugPrintMsg(x)

#endif

void DebugPrint2(int max, const char* format, ...);
void DebugPrint(const char* format, ...);

#ifdef __cplusplus
}
#endif

#endif // _FILTDEBUG__H
