<!DOCTYPE html>
<html>
	<head>
		<!-- Required meta tags -->
		<meta charset="utf-8">
		<meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">

		<!-- Bootstrap CSS -->
		<link rel="stylesheet" href="{{ url_for('static', filename='bootstrap.css') }}">
	</head>

	<body>
		<div class="container">
			{% block content %}{% endblock %}
		</div>
	</body>
</html>
